import * as plugins from './plugins.js';
import * as interfaces from '../dist_ts_interfaces/index.js';
import { logger } from './logging.js';

// Add type definitions for ServiceWorker APIs
declare global {
  interface ServiceWorkerGlobalScope extends EventTarget {
    clients: Clients;
    registration: ServiceWorkerRegistration;
  }
  
  // Define Clients interface
  interface Clients {
    matchAll(options?: ClientQueryOptions): Promise<Client[]>;
    openWindow(url: string): Promise<WindowClient>;
    claim(): Promise<void>;
    get(id: string): Promise<Client | undefined>;
  }
  
  interface ClientQueryOptions {
    includeUncontrolled?: boolean;
    type?: 'window' | 'worker' | 'sharedworker' | 'all';
  }
  
  interface Client {
    id: string;
    type: 'window' | 'worker' | 'sharedworker';
    url: string;
  }
  
  interface WindowClient extends Client {
    focused: boolean;
    visibilityState: 'hidden' | 'visible' | 'prerender' | 'unloaded';
    focus(): Promise<WindowClient>;
    navigate(url: string): Promise<WindowClient>;
  }
}

/**
 * This class is meant to be used only on the backend side
 */
export class ServiceworkerBackend {
  public deesComms = new plugins.deesComms.DeesComms();

  constructor(optionsArg: {
    self: any;
    purgeCache: (reqArg: interfaces.serviceworker.IRequest_PurgeServiceWorkerCache['request']) => Promise<interfaces.serviceworker.IRequest_PurgeServiceWorkerCache['response']>;
  }) {

    // lets handle wakestuff
    optionsArg.self.addEventListener('message', (event) => {
      if (event.data && event.data.type === 'wakeUpCall') {
        console.log('sw-backend: got wake up call');
      }
    });
    this.deesComms.createTypedHandler<interfaces.serviceworker.IRequest_Client_Serviceworker_ConnectionPolling>('broadcastConnectionPolling', async reqArg => {
      return {
        serviceworkerId: '123'
      };
    })

    this.deesComms.createTypedHandler<interfaces.serviceworker.IRequest_PurgeServiceWorkerCache>('purgeServiceWorkerCache', async reqArg => {
      console.log(`Executing purge cache in serviceworker backend.`)
      return await optionsArg.purgeCache?.(reqArg);
    });
  }

  /**
   * reloads all clients
   */
  public async triggerReloadAll() {
    try {
      logger.log('info', 'Triggering reload for all clients due to new version');
      
      // Send update message via DeesComms
      // This will be picked up by clients that have registered a handler for 'serviceworker_newVersion'
      await this.deesComms.postMessage({
        method: 'serviceworker_newVersion',
        request: {},
        messageId: `sw_update_${Date.now()}`
      });
      
      // As a fallback, also use the clients API to reload clients that might not catch the broadcast
      // We need to type-cast self since TypeScript doesn't recognize ServiceWorker API
      const swSelf = self as unknown as ServiceWorkerGlobalScope;
      const clients = await swSelf.clients.matchAll({ type: 'window' });
      logger.log('info', `Found ${clients.length} clients to reload`);
      
      for (const client of clients) {
        if ('navigate' in client) {
          // For modern browsers, navigate to the same URL to trigger reload
          (client as any).navigate(client.url);
          logger.log('info', `Navigated client to: ${client.url}`);
        }
      }
    } catch (error) {
      logger.log('error', `Failed to reload clients: ${error instanceof Error ? error.message : String(error)}`);
    }
  }

  /**
   * display notification
   */
  public async addNotification(notificationArg: {
    title: string;
    body: string;
  }) {
    try {
      // Check if we have permission to show notifications
      const permission = self.Notification?.permission;
      if (permission !== 'granted') {
        logger.log('warn', `Cannot show notification: permission is ${permission}`);
        return;
      }
      
      // Type-cast self to ServiceWorkerGlobalScope
      const swSelf = self as unknown as ServiceWorkerGlobalScope;
      
      // Use type assertion for notification options to include vibrate
      const options = {
        body: notificationArg.body,
        icon: '/favicon.ico', // Assuming there's a favicon
        badge: '/favicon.ico',
        vibrate: [200, 100, 200]
      } as NotificationOptions;
      
      await swSelf.registration.showNotification(notificationArg.title, options);
      
      logger.log('info', `Notification shown: ${notificationArg.title}`);
    } catch (error) {
      logger.log('error', `Failed to show notification: ${error instanceof Error ? error.message : String(error)}`);
    }
  }

  public async alert(alertText: string) {
    // Since we can't directly show alerts from service worker context,
    // we'll use notifications as a fallback
    await this.addNotification({
      title: 'Alert',
      body: alertText
    });
    
    // Send message to clients who might be able to show an actual alert
    try {
      await this.deesComms.postMessage({
        method: 'serviceworker_alert',
        request: { message: alertText },
        messageId: `sw_alert_${Date.now()}`
      });
      logger.log('info', `Alert message sent to clients: ${alertText}`);
    } catch (error) {
      logger.log('error', `Failed to send alert to clients: ${error instanceof Error ? error.message : String(error)}`);
    }
  }
}